class ExpiredError(Exception):
    """
    Raise when item is expired.
    """

class InvalidKeyError(Exception):
    """
    Raise when trying to authenticate with an unknown/invalid key.
    """

class InvalidTokenError(Exception):
    """
    Raise when trying to use an unknown/invalid token.
    """

class RefreshError(Exception):
    """
    Raise when refreshing JWT fails.
    """

class AuthenticationError(Exception):
    """
    Raise when failed to authenticate.
    """

class MissingCredentialsError(Exception):
    """
    Raise when no credentials were given to Client class.
    """

class InvalidServerURLError(Exception):
    """
    Raised if an invalid server URL is given.
    """

class RegisterKeyError(Exception):
    """
    Raised when failed to register key to server.
    """

class IsKeyRegisteredError(Exception):
    """
    Raised when failed to check if key is registered.
    """

class MissingAuthorizationError(Exception):
    """
    Raised when authorization header is missing from request.
    """

class BadAuthorizationError(Exception):
    """
    Raised when authorization header is incorrect.
    """

class MissingJWTError(Exception):
    """
    Raised if JWT is misisng.
    """



import asyncio
from jwthenticator.consts import DB_URI

main_event_loop = asyncio.new_event_loop()
db_lock = asyncio.Lock(loop=main_event_loop)

def is_using_sqlite()->bool:
    return "sqlite://" in DB_URI



from __future__ import absolute_import

from typing import Optional
from datetime import datetime, timedelta
from hashlib import sha512
from uuid import UUID

from sqlalchemy import select, func

from jwthenticator.utils import create_async_session_factory
from jwthenticator.schemas import KeyData
from jwthenticator.models import Base, KeyInfo
from jwthenticator.exceptions import InvalidKeyError
from jwthenticator.consts import KEY_EXPIRY, ASYNC_DB_URI


class KeyManager:
    """
    Class used for creation and loading of keys.
    """

    def __init__(self) -> None:
        self.async_session_factory = create_async_session_factory(ASYNC_DB_URI, Base)
        self.key_schema = KeyData.Schema()


    async def create_key(self, key: str, identifier: UUID, expires_at: Optional[datetime] = None) -> bool:
        """
        Add a new key to database.
        Will update the key's expiry date if key already exists.
        :param key: The key to register.
        :return: Returns True if successfull, raises exception otherwise.
        """
        if expires_at is None:
            expires_at = datetime.utcnow() + timedelta(seconds=KEY_EXPIRY)
        key_hash = sha512(key.encode()).hexdigest()

        # If key already exists, update expiry date.
        if await self.check_key_exists(key_hash):
            return await self.update_key_expiry(key_hash, expires_at)

        key_obj = KeyInfo(
            expires_at=expires_at,
            key_hash=key_hash,
            identifier=identifier
        )
        async with self.async_session_factory() as session:
            async with session.begin():
                session.add(key_obj)
        return True


    async def check_key_exists(self, key_hash: str) -> bool:
        """
        Check if a key exists in DB.
        """
        async with self.async_session_factory() as session:
            query = select(func.count(KeyInfo.id)).where(KeyInfo.key_hash == key_hash)
            return (await session.scalar(query)) == 1


    async def update_key_expiry(self, key_hash: str, expires_at: datetime) -> bool:
        """
        Update the expiry date of an existing key.
        """
        if not await self.check_key_exists(key_hash):
            raise InvalidKeyError("Invalid key")
        async with self.async_session_factory() as session:
            query = select(KeyInfo).where(KeyInfo.key_hash == key_hash)
            key_info_obj = await session.scalar(query)
            key_info_obj.expires_at = expires_at
        return True


    async def get_key(self, key_hash: str) -> KeyData:
        """
        Load key data from database.
        """
        if not await self.check_key_exists(key_hash):
            raise InvalidKeyError("Invalid key")
        async with self.async_session_factory() as session:
            query = select(KeyInfo).where(KeyInfo.key_hash == key_hash)
            key_info_obj = await session.scalar(query)
            key_data_obj = self.key_schema.load((self.key_schema.dump(key_info_obj)))
            return key_data_obj



# pylint: disable=invalid-name
from __future__ import absolute_import

import uuid
from dataclasses import field
from typing import Optional, List, ClassVar, Type, Dict, Any
from datetime import datetime
from collections.abc import Hashable

from marshmallow import Schema, fields, post_dump
from marshmallow_dataclass import dataclass, NewType, add_schema

from jwthenticator.consts import JWT_ALGORITHM, JWT_ALGORITHM_FAMILY

# Define the UUID type that uses Marshmallow's UUID + Python's UUID
UUID = NewType("UUID", uuid.UUID, field=fields.UUID)


# Marshmallow base schema for skipping None values on dump
class BaseSchema(Schema):
    SKIP_VALUES = {None}

    @post_dump
    # pylint: disable=unused-argument
    def remove_skip_values(self, data: Any, many: bool) -> Dict[Any, Any]:
        return {
            key: value for key, value in data.items()
            if not isinstance(value, Hashable) or value not in self.SKIP_VALUES
        }


# Data dataclasses (that match the sqlalchemy models)
@dataclass  # pylint: disable=used-before-assignment
class KeyData:
    Schema: ClassVar[Type[Schema]] = Schema
    id: int
    created: datetime
    expires_at: datetime
    key_hash: str
    key: Optional[str] = field(default=None, repr=False, metadata={"load_only": True})

    async def is_valid(self) -> bool:
        return self.expires_at > datetime.utcnow()


@dataclass
class RefreshTokenData:
    Schema: ClassVar[Type[Schema]] = Schema
    id: int
    created: datetime
    expires_at: datetime
    token: str
    key_id: int

    async def is_valid(self) -> bool:
        return self.expires_at > datetime.utcnow()


# Skipping None values on dump since 'aud' is optional and can't be None/empty
@add_schema(base_schema=BaseSchema)
@dataclass
class JWTPayloadData:
    Schema: ClassVar[Type[Schema]] = Schema
    token_id: UUID   # JWT token identifier
    identifier: UUID # Identifier of who the JWT was issued to
    iat: int    # Issued at timestamp
    exp: int    # Expires at timestamp
    aud: Optional[List[str]] = None   # JWT Audience

    async def is_valid(self) -> bool:
        return self.exp > datetime.utcnow().timestamp()


# Request dataclasses
@dataclass
class KeyRequest:
    Schema: ClassVar[Type[Schema]] = Schema
    key: str


@dataclass
class RegisterKeyRequest(KeyRequest):
    identifier: UUID # Identifier of who the JWT was issued to


@dataclass
class RefreshRequest:
    Schema: ClassVar[Type[Schema]] = Schema
    refresh_token: str
    identifier: UUID


@dataclass
class JWTValidateRequest:
    Schema: ClassVar[Type[Schema]] = Schema
    jwt: str


@dataclass
class AuthRequest:
    Schema: ClassVar[Type[Schema]] = Schema
    key: str = field(repr=False)
    identifier: UUID


# Response dataclasses
@dataclass
class TokenResponse:
    Schema: ClassVar[Type[Schema]] = Schema
    jwt: str
    refresh_token: Optional[str] = None


@dataclass
class BoolResponse:
    Schema: ClassVar[Type[Schema]] = Schema
    result: Optional[bool]
    message: Optional[str] = None


@dataclass
class JWKPayload:   # pylint: disable=too-many-instance-attributes
    """
    See https://auth0.com/docs/tokens/json-web-tokens/json-web-key-set-properties
    """
    Schema: ClassVar[Type[Schema]] = Schema
    x5c: List[str]  # x.509 certificate chain
    n: bytes    # RSA public key modulus
    e: bytes    # RSA public key exponent
    x5t: str    # x.509 SHA-1 thumbprint
    kid: str    # Unique key identifier
    alg: str = JWT_ALGORITHM    # Key algorithm
    kty: str = JWT_ALGORITHM_FAMILY # Key algorithm family
    use: str = "sig"    # How key will be used, sig (signature) by default


@dataclass
class JWKSResponse:
    Schema: ClassVar[Type[Schema]] = Schema
    keys: List[JWKPayload]



from __future__ import absolute_import

import asyncio
from os.path import isfile
from typing import Any, Dict, Tuple, Optional
from urllib.parse import urlparse

from jwt.utils import base64url_encode
from Cryptodome.PublicKey import RSA
from Cryptodome.Hash import SHA1
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession, AsyncEngine
from sqlalchemy.pool import NullPool

from jwthenticator.consts import RSA_KEY_STRENGTH, RSA_PUBLIC_KEY, RSA_PRIVATE_KEY, RSA_PUBLIC_KEY_PATH, RSA_PRIVATE_KEY_PATH


def get_rsa_key_pair() -> Tuple[str, Optional[str]]:
    """
    Get RSA key pair.
    Will get RSA key pair depending on available ENV variables, in the following order:
    1. Read file path from RSA_PUBLIC_PATH and PRIVATE_KEY_PATH and use the files there,
       If a path is specified (in RSA_PUBLIC_PATH and PRIVATE_KEY_PATH) but the files do
       not exist - they will be created and populated
    2. Use data directly in the env vars RSA_PUBLIC_KEY and RSA_PRIVATE_KEY
    3. Use stateless new keys
    :return (public_key, private_key): A key pair tuple.
        Will raise exception if key paths are given and fail to read.
    """
    if RSA_PUBLIC_KEY_PATH:
        if isfile(RSA_PUBLIC_KEY_PATH):
            return _read_rsa_keys_from_file()
        return _create_rsa_key_files()

    if RSA_PUBLIC_KEY:
        return RSA_PUBLIC_KEY, RSA_PRIVATE_KEY

    return create_rsa_key_pair()


def _read_rsa_keys_from_file() -> Tuple[str, Optional[str]]:
    with open(RSA_PUBLIC_KEY_PATH, 'r', encoding='utf8') as f_obj:
        public_key = f_obj.read()
    private_key = None
    if RSA_PRIVATE_KEY_PATH is not None:
        with open(RSA_PRIVATE_KEY_PATH, 'r', encoding='utf8') as f_obj:
            private_key = f_obj.read()
    return public_key, private_key


def _create_rsa_key_files() -> Tuple[str, Optional[str]]:
    public_key, private_key = create_rsa_key_pair()
    with open(RSA_PUBLIC_KEY_PATH, 'w', encoding='utf8') as f_obj:
        f_obj.write(public_key)
    with open(RSA_PRIVATE_KEY_PATH, 'w', encoding='utf8') as f_obj:
        f_obj.write(private_key)
    return public_key, private_key


def create_rsa_key_pair() -> Tuple[str, str]:
    """
    Create RSA key pair.
    Function is sync so it can be used from __init__ func.
    :return (public_key, private_key): A key pair tuple.
    """
    key = RSA.generate(RSA_KEY_STRENGTH)
    public_key = key.publickey().export_key().decode()
    private_key = key.export_key().decode()
    return public_key, private_key


def calculate_key_signature(public_key: str) -> str:
    """
    Calculate the SHA1 signature for a given RSA public key.
    Calculation follows the JWK RFC7517 section 4.8 meaning base64
        URL encoding of SHA1 signature of public RSA key.
    :param public_key: The public key to calculate signature for.
    """
    rsa_obj = RSA.import_key(public_key)
    rsa_der = rsa_obj.export_key("DER")

    hasher = SHA1.new()
    hasher.update(rsa_der)
    fingerprint = base64url_encode(hasher.digest())

    return fingerprint.decode("utf8")


def verify_url(url: str) -> bool:
    """
    Verifies a given URL is valid and will be accepted by http client.
    """
    parsed_url = urlparse(url)
    return all([parsed_url.scheme, parsed_url.netloc])


def fix_url_path(url: str) -> str:
    """
    Add "/" to end of URL, if URL has path and doesn't end with "/"
    the path will be removed by urljoin.
    """
    return url if url.endswith("/") else url + "/"


async def create_base(engine: AsyncEngine, base: DeclarativeMeta) -> None:
    async with engine.begin() as conn:
        await conn.run_sync(base.metadata.create_all)


def create_async_session_factory(uri: str, base: Optional[DeclarativeMeta] = None, **engine_kwargs: Dict[Any, Any]) -> sessionmaker:
    """
    :param uri: Database uniform resource identifier
    :param base: Declarative SQLAlchemy class to base off table initialization
    :param engine_kwargs: Arguments to pass to SQLAlchemy's engine initialization
    :returns: :class:`.AsyncSession` factory
    """
    engine = create_async_engine(uri, **engine_kwargs, poolclass=NullPool)
    if base is not None:
        asyncio.get_event_loop().run_until_complete(create_base(engine, base))
    return sessionmaker(engine, expire_on_commit=False, class_=AsyncSession)



# pylint: disable=too-few-public-methods
from __future__ import absolute_import

from typing import Any
from datetime import datetime

from sqlalchemy import create_engine, Column, Integer, String, DateTime, ForeignKey
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy_utils import database_exists, create_database
from sqlalchemy_utils.types.uuid import UUIDType

from jwthenticator.consts import DB_URI

engine = create_engine(DB_URI)
SessionMaker = sessionmaker(bind=engine)

Base = declarative_base()   # type: Any # pylint: disable=invalid-name


class KeyInfo(Base):
    __tablename__ = "keys"
    id = Column(Integer, primary_key=True, autoincrement=True)
    created = Column(DateTime, default=datetime.utcnow())
    expires_at = Column(DateTime)
    key_hash = Column(String(256), unique=True)
    identifier = Column(UUIDType(binary=False), nullable=False)


class RefreshTokenInfo(Base):
    __tablename__ = "refresh_tokens"
    id = Column(Integer, primary_key=True, autoincrement=True)
    created = Column(DateTime, default=datetime.utcnow())
    expires_at = Column(DateTime)
    token = Column(String(512))
    key_id = Column(Integer, ForeignKey("keys.id"))


# Create database + tables
if not database_exists(DB_URI):
    create_database(DB_URI)
Base.metadata.create_all(engine)



from __future__ import absolute_import

from functools import wraps
from typing import Any, Dict, Callable
from urllib.parse import urljoin
from http import HTTPStatus

from aiohttp import ClientSession, web

from jwthenticator.schemas import JWTValidateRequest
from jwthenticator.utils import verify_url, fix_url_path
from jwthenticator.exceptions import InvalidServerURLError, MissingAuthorizationError, BadAuthorizationError, MissingJWTError


def authenticate(jwthenticator_server: str) -> Any:
    """
    Wrapper function to verify user performing the request is authorized.
    The function extracts the JWT token from the header and verifies it against
        an jwthenticator server (given by jwthenticator_server).
    Use it to wrap async endpoints that require authentication.
    :param jwthenticator_server: The (full) URL of the jwthenticator server.
        For example - http://localhost.
    """
    def wrap(func: Callable) -> Any:
        @wraps(func)
        async def async_wrap(request: web.Request, *args: Any, **kwargs: Dict[Any, Any]) -> web.Response:
            # Verify jwthenticator_server is a proper URL
            if not verify_url(jwthenticator_server):
                raise InvalidServerURLError()
            valid_jwthenticator_server = fix_url_path(jwthenticator_server)

            # Extract JWT
            try:
                jwt = await extract_jwt(request)
            except MissingAuthorizationError:
                return web.json_response({}, status=HTTPStatus.FORBIDDEN)
            except BadAuthorizationError:
                return web.json_response({}, status=HTTPStatus.BAD_REQUEST)

            # Verify JWT
            if not await verify_jwt(valid_jwthenticator_server, jwt):
                return web.json_response({}, status=HTTPStatus.UNAUTHORIZED)

            # Verified! Perform actual call
            return await func(request, *args, **kwargs)

        return async_wrap
    return wrap


async def verify_jwt(jwthenticator_server: str, jwt: str) -> bool:
    """
    Verify agains jwthenticator server that the given JWT is valid.
    :return bool: True if valid, False if isn't.
    """
    if not jwthenticator_server.endswith("/"):
        jwthenticator_server = jwthenticator_server + "/"
    url = urljoin(jwthenticator_server, "validate")
    request = JWTValidateRequest(jwt)
    async with ClientSession() as client:
        async with client.post(url, json=JWTValidateRequest.Schema().dump(request)) as response:
            return response.status == HTTPStatus.OK


async def extract_jwt(request: web.Request) -> str:
    """
    Extract JWT from request header.
    :param request: Request to extract JWT header from.
    :return: JWT if exists, raises exception otherwise.
    """
    # Extract authorization header.
    authorization = request.headers.get("Authorization", None)
    if authorization is None:
        raise MissingAuthorizationError

    # Extract JWT from header
    split_authorization = authorization.split(" ")
    if len(split_authorization) != 2 or split_authorization[0] != "Bearer":
        raise BadAuthorizationError

    jwt = split_authorization[1]
    if not jwt:
        raise MissingJWTError
    return jwt



from __future__ import absolute_import

from environs import Env

def days_to_seconds(days: int) -> int:
    return days * 60 * 24

env = Env()


# Server consts
PORT = env.int("PORT", 8080)
URL_PREFIX = URL_PREFIX = env.str("URL_PREFIX", "")
DISABLE_EXTERNAL_API = env.bool("DISABLE_EXTERNAL_API", False)
DISABLE_INTERNAL_API = env.bool("DISABLE_INTERNAL_API", False)

# JWT consts
JWT_ALGORITHM = env.str("JWT_ALGORITHM", "RS256")
JWT_ALGORITHM_FAMILY = env.str("JWT_ALGORITHM_FAMILY", "RSA")
JWT_LEASE_TIME = env.int("JWT_LEASE_TIME", 25 * 3600) # In seconnds - 25 hours to allow for client time zone issues
JWT_AUDIENCE = env.list("JWT_AUDIENCE", [])

# Token consts
KEY_EXPIRY = env.int("KEY_EXPIRY", days_to_seconds(120))  # In seconds
REFRESH_TOKEN_EXPIRY = env.int("REFRESH_TOKEN_EXPIRY", days_to_seconds(60))   # In seconds

# Keys from env
RSA_PUBLIC_KEY = env("RSA_PUBLIC_KEY", None)
RSA_PRIVATE_KEY = env("RSA_PRIVATE_KEY", None)
# Keys from file
RSA_PUBLIC_KEY_PATH = env("RSA_PUBLIC_KEY_PATH", None)
RSA_PRIVATE_KEY_PATH = env("RSA_PRIVATE_KEY_PATH", None)
# If no key is given, used for key generation
RSA_KEY_STRENGTH = env.int("RSA_KEY_SIZE", 2048)

# DB consts
DB_CONNECTOR = env.str("DB_CONNECTOR", "postgresql+pg8000")
ASYNC_DB_CONNECTOR = env.str("ASYNC_DB_CONNECTOR", "postgresql+asyncpg")
DB_USER = env.str("DB_USER", "postgres")
DB_PASS = env.str("DB_PASS", "")
DB_HOST = env.str("DB_HOST", "localhost")
DB_NAME = env.str("DB_NAME", "jwthenticator")

DB_URI = env.str("DB_URI", f"{DB_CONNECTOR}://{DB_USER}:{DB_PASS}@{DB_HOST}/{DB_NAME}")
ASYNC_DB_URI = env.str("ASYNC_URI", f"{ASYNC_DB_CONNECTOR}://{DB_USER}:{DB_PASS}@{DB_HOST}/{DB_NAME}")



from __future__ import absolute_import

from base64 import b64encode
from hashlib import sha512
from typing import Optional, Tuple, List

from Cryptodome.PublicKey import RSA
from jwt.utils import to_base64url_uint

from jwthenticator import schemas
from jwthenticator.tokens import TokenManager
from jwthenticator.keys import KeyManager
from jwthenticator.consts import JWT_ALGORITHM, JWT_ALGORITHM_FAMILY, JWT_LEASE_TIME, JWT_AUDIENCE
from jwthenticator.utils import get_rsa_key_pair, calculate_key_signature
from jwthenticator.exceptions import ExpiredError
from jwthenticator.loop_management import is_using_sqlite, db_lock


class JWThenticatorAPI:
    """
    High level authentication API functions.
    Class uses TokenManager + KeyManager and provides API logic using them.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, rsa_key_pair: Tuple[str, Optional[str]] = get_rsa_key_pair(),
                 jwt_lease_time: int = JWT_LEASE_TIME, jwt_algorithm: str = JWT_ALGORITHM,
                 jwt_algorithm_family: str = JWT_ALGORITHM_FAMILY, jwt_audience: List[str] = JWT_AUDIENCE):
        """
        Class can be initiated without giving any parameter, will generate RSA key pair by itself.
        :param rsa_key_pair: (public_key, private_key) RSA key pair. Will generate keys if not given
            keys or paths through consts. In case private_key is None will only support token validation.
        """
        self.public_key, self._private_key = rsa_key_pair
        self.jwt_algorithm = jwt_algorithm
        self.jwt_algorithm_family = jwt_algorithm_family
        self.key_signature = calculate_key_signature(self.public_key)

        self.token_manager = TokenManager(self.public_key, self._private_key, self.jwt_algorithm, jwt_lease_time, jwt_audience, key_id=self.key_signature)
        self.key_manager = KeyManager()


    async def authenticate(self, request: schemas.AuthRequest) -> schemas.TokenResponse:
        """
        Authenticate using a key.
        :return: access + refresh tokens if successfull
        """
        key_hash = sha512(request.key.encode()).hexdigest()
        key_obj = await self.key_manager.get_key(key_hash)
        if not await key_obj.is_valid():
            raise ExpiredError("Key is expired.")

        jwt_token = await self.token_manager.create_access_token(request.identifier)

        if is_using_sqlite():
            # this lock is designed to block multiple request to sqlite causing DB lock
            async with db_lock:
                refresh_token = await self.token_manager.create_refresh_token(key_obj.id)
        else:
            refresh_token = await self.token_manager.create_refresh_token(key_obj.id)

        return schemas.TokenResponse(
            jwt=jwt_token,
            refresh_token=refresh_token
        )


    async def refresh(self, request: schemas.RefreshRequest) -> schemas.TokenResponse:
        """
        Use a refresh token to create a new access token.
        """
        refresh_token_obj = await self.token_manager.load_refresh_token(request.refresh_token)
        if not await refresh_token_obj.is_valid():
            raise ExpiredError("Token is expired.")

        jwt_token = await self.token_manager.create_access_token(request.identifier)
        return schemas.TokenResponse(jwt=jwt_token)


    async def validate(self, request: schemas.JWTValidateRequest) -> schemas.BoolResponse:
        """
        Validate a JWT token.
        Raises a %s exception if token is invalid.
        """
        token = await self.token_manager.load_access_token(request.jwt)
        if not await token.is_valid():
            raise ExpiredError("Token is expired.")
        return schemas.BoolResponse(result=True)


    async def register_key(self, request: schemas.RegisterKeyRequest) -> schemas.BoolResponse:
        """
        Register a new key so it can be used for authentication.
        """
        await self.key_manager.create_key(request.key, request.identifier)
        return schemas.BoolResponse(result=True)


    async def is_key_registerd(self, request: schemas.KeyRequest) -> schemas.BoolResponse:
        """
        Check if a given key is already registered.
        """
        key_hash = sha512(request.key.encode()).hexdigest()
        result = await self.key_manager.check_key_exists(key_hash)
        return schemas.BoolResponse(result=result)


    async def get_jwks(self) -> schemas.JWKSResponse:
        """
        Get the JWKS requried for authentication.
        See here for more details: https://auth0.com/docs/tokens/references/jwks-properties
        """
        rsa_obj = RSA.import_key(self.public_key)
        rsa_der = rsa_obj.export_key("DER")

        jwk_payload = schemas.JWKPayload(
            alg=self.jwt_algorithm,
            kty=self.jwt_algorithm_family,
            use="sig",
            x5c=[b64encode(rsa_der).decode("utf8")],
            n=to_base64url_uint(rsa_obj.n),
            e=to_base64url_uint(rsa_obj.e),
            kid=self.key_signature,
            x5t=self.key_signature
        )

        jwks_obj = schemas.JWKSResponse([jwk_payload])
        return jwks_obj



from __future__ import absolute_import

from typing import Optional, Any, Dict
from datetime import datetime
from urllib.parse import urljoin
from http import HTTPStatus
from uuid import UUID

import jwt as pyjwt # To avoid redfinition in class
from aiohttp import ClientSession, ClientResponse

from jwthenticator import schemas, exceptions
from jwthenticator.utils import verify_url, fix_url_path
from jwthenticator.consts import JWT_ALGORITHM

JWT_DECODE_OPTIONS = {"verify_signature": False, "verify_exp": False}


# pylint: disable=too-many-instance-attributes
class Client:
    """
    Class provides state management for working with JWThenticator.
    Exporting wrapper functions for "no brainer" usage against JWThenticator.
    All functions in this class are async.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, jwthenticator_server: str, identifier: UUID, jwt: Optional[str] = None,
                 refresh_token: Optional[str] = None, key: Optional[str] = None,
                 verify_ssl: Optional[bool] = None, algorithm: str = JWT_ALGORITHM) -> None:
        """
        :param jwthenticator_server: The (full) URL of the jwthenticator server.
            For example - http://localhost:8080/.
        :param identifier: The UUID identifier that will be used for all operations against
            jwthenticator server.
        :param jwt: JWT (if you already have one previsioned).
        :param refresh_token: Refresh token (if you already have one previsioned).
        :param key: Key to used to authenticate against jwthenticator server.
        :param verify_ssl: The SSL validation mode given to aiohttp client. Use None for default
            one or False to disable certificate check.
        :param algorithm: The JWT algorithm used. This is needed for decoding received JWTs
            and extracting their expiry time. If not given, default will be used.
        """
        self.jwthenticator_server = jwthenticator_server
        self.identifier = identifier
        self.verify_ssl = verify_ssl
        self.algorithm = algorithm

        self._jwt_exp = None
        self.jwt = jwt
        self._refresh_token = refresh_token
        self._key = key

        # Verify jwthenticator_server is a proper URL
        if not verify_url(self.jwthenticator_server):
            raise exceptions.InvalidServerURLError()
        self.jwthenticator_server = fix_url_path(self.jwthenticator_server)

        # Verify at least one of these was given.
        if self.jwt is None and self._refresh_token is None and self._key is None:
            raise exceptions.MissingCredentialsError()

        self.auth_request_schema = schemas.AuthRequest.Schema()
        self.refresh_request_schema = schemas.RefreshRequest.Schema()
        self.token_response_schema = schemas.TokenResponse.Schema()


    @property
    def jwt(self) -> Optional[str]:
        return self._jwt

    @jwt.setter
    def jwt(self, value: str) -> None:
        self._jwt = value
        if value:
            # Algorithm is given (even though the client doesn't care) since it's required by pyjwt.
            self._jwt_exp = pyjwt.decode(value, options=JWT_DECODE_OPTIONS, algorithms=[self.algorithm]).get("exp")
        else:
            self._jwt_exp = None

    @property
    def is_jwt_expired(self) -> bool:
        if self._jwt_exp is None:
            return True
        return datetime.utcnow().timestamp() >= self._jwt_exp

    @property
    def refresh_token(self) -> Optional[str]:
        return self._refresh_token

    @property
    def header(self) -> Dict[str, str]:
        """
        Get requests style header with JWT.
        """
        return {"Authorization": f"Bearer {self.jwt}"}


    async def refresh(self, auth_on_fail: bool = True) -> None:
        """
        Perfrom "refresh" request.
        :param auth_on_fail: Whether to try and authenticate if refreshing fails.
        :return: None or raises exception if fails.
        """
        if self._refresh_token is None:
            return await self.authenticate()

        url = urljoin(self.jwthenticator_server, "refresh")
        request = schemas.RefreshRequest(self._refresh_token, self.identifier)
        async with ClientSession() as client:
            async with client.post(url, json=self.refresh_request_schema.dump(request), ssl=self.verify_ssl) as response:

                if response.status != HTTPStatus.OK:
                    if response.status == HTTPStatus.UNAUTHORIZED:
                        self._refresh_token = None
                    if auth_on_fail:
                        return await self.authenticate()
                    raise exceptions.RefreshError()

                token_response_data = await response.json()

        result = self.token_response_schema.load(token_response_data)
        self.jwt = result.jwt


    async def authenticate(self) -> None:
        """
        Perform "authenticate" request.
        :return: None or raises exception if fails.
        """
        if self._key is None:
            raise TypeError("Missing key")

        url = urljoin(self.jwthenticator_server, "authenticate")
        request = schemas.AuthRequest(self._key, self.identifier)
        async with ClientSession() as client:
            async with client.post(url, json=self.auth_request_schema.dump(request), ssl=self.verify_ssl) as response:

                if response.status != HTTPStatus.OK:
                    if response.status == HTTPStatus.UNAUTHORIZED:
                        self._key = None
                    raise exceptions.AuthenticationError()

                token_response_data = await response.json()

        result = self.token_response_schema.load(token_response_data)
        self.jwt = result.jwt
        self._refresh_token = result.refresh_token if result.refresh_token else self._refresh_token


    async def _make_request(self, method: str, url: str, **kwargs: Dict[str, Any]) -> ClientResponse:
        """
        Add auth headers and make actually make the request.
        """
        headers = kwargs.get("headers", {})
        headers.update(self.header)
        kwargs["headers"] = headers
        async with ClientSession() as client:
            response = await client.request(method, url, ssl=self.verify_ssl, **kwargs)
        return response


    async def request_with_auth(self, method: str, url: str, **kwargs: Dict[str, Any]) -> ClientResponse:
        """
        Perform request with authentication headers.
        """
        # Refresh if JWT is expired or doesn't exist, done to minimize number of requests to server.
        if self.is_jwt_expired or self.jwt is None:
            await self.refresh()

        # Try to make request
        response = await self._make_request(method, url, **kwargs)

        # If receive unauthorized, try to refresh once more (possibly expired or invalid JWT).
        if response.status == HTTPStatus.UNAUTHORIZED:
            self.jwt = None
            await self.refresh()
            response = await self._make_request(method, url, **kwargs)

        return response


    async def get_with_auth(self, url: str, **kwargs: Dict[str, Any]) -> ClientResponse:
        """
        Perform get request with JWT header.
        """
        return await self.request_with_auth("GET", url, **kwargs)


    async def post_with_auth(self, url: str, **kwargs: Dict[str, Any]) -> ClientResponse:
        """
        Perform post request with JWT header.
        """
        return await self.request_with_auth("POST", url, **kwargs)



class InternalClient:
    """
    Class provides easier calling to the JWThenticator's "internal" functionality.
    All function in this class are async.
    """

    def __init__(self, jwthenticator_server: str, identifier: UUID, verify_ssl: Optional[bool] = None) -> None:
        """
        :param jwthenticator_server: The (full) URL of the jwthenticator server.
            For example - http://localhost:8080/.
        :param identifier: The UUID identifier that will be used for all operations against
            jwthenticator server.
        :param verify_ssl: The SSL validation mode given to aiohttp client. Use None for default
            one or False to disable certificate check.
        """
        self.jwthenticator_server = jwthenticator_server
        self.identifier = identifier
        self.verify_ssl = verify_ssl

        # Verify jwthenticator_server is a proper URL
        if not verify_url(self.jwthenticator_server):
            raise exceptions.InvalidServerURLError()
        self.jwthenticator_server = fix_url_path(self.jwthenticator_server)

        self.register_key_request_schema = schemas.RegisterKeyRequest.Schema()
        self.key_request_schema = schemas.KeyRequest.Schema()
        self.bool_response_schema = schemas.BoolResponse.Schema()


    async def register_key(self, key: str) -> None:
        """
        Register a key to the jwthenticator server.
        :param key: The key to register.
        :return: None or raises exception if fails.
        """
        url = urljoin(self.jwthenticator_server, "register_key")
        request = schemas.RegisterKeyRequest(key, self.identifier)
        async with ClientSession() as client:
            async with client.post(url, json=self.register_key_request_schema.dump(request), ssl=self.verify_ssl) as response:

                if response.status != HTTPStatus.CREATED:
                    raise exceptions.RegisterKeyError(response.status)

                bool_response_data = await response.json()

        result = self.bool_response_schema.load(bool_response_data)
        if not result.result:
            raise exceptions.RegisterKeyError(result.message)


    async def is_key_registered(self, key: str) -> bool:
        """
        Check if a key is already registered in the jwthenticator server.
        :param key: The key to check if registered.
        :return: None or raises exception if fails.
        """
        url = urljoin(self.jwthenticator_server, "is_key_registered")
        request = schemas.KeyRequest(key)
        async with ClientSession() as client:
            async with client.post(url, json=self.key_request_schema.dump(request), ssl=self.verify_ssl) as response:

                if response.status != HTTPStatus.OK:
                    raise exceptions.IsKeyRegisteredError(response.status)

                bool_response_data = await response.json()

        result = self.bool_response_schema.load(bool_response_data)
        return result.result



from __future__ import absolute_import

from typing import Optional, List
from datetime import datetime, timedelta
from hashlib import sha512
from uuid import UUID, uuid4

import jwt
from sqlalchemy import select, func

from jwthenticator.utils import create_async_session_factory
from jwthenticator.models import Base, RefreshTokenInfo
from jwthenticator.schemas import JWTPayloadData, RefreshTokenData
from jwthenticator.exceptions import InvalidTokenError, MissingJWTError
from jwthenticator.consts import JWT_ALGORITHM, REFRESH_TOKEN_EXPIRY, JWT_LEASE_TIME, JWT_AUDIENCE, ASYNC_DB_URI

class TokenManager:
    """
    Class responsible for the creation and loading of tokens
    """

    # pylint: disable=too-many-arguments,too-many-instance-attributes
    def __init__(self, public_key: str, private_key: Optional[str] = None, algorithm: str = JWT_ALGORITHM,
                 jwt_lease_time: int = JWT_LEASE_TIME, jwt_audience: List[str] = JWT_AUDIENCE, key_id: Optional[str] = None):
        """
        Accepts public + private key pairs.
        If only public key is given tokens can be loaded but not created.
        """
        # This is done to avoid an "exploit" where algorithm=none leaving the jwt tokens unsecure.
        if not algorithm or algorithm.lower() == "none":
            raise ValueError("Algorithm can't be empty!")

        self.public_key = public_key
        self.private_key = private_key
        self.algorithm = algorithm
        self.jwt_lease_time = jwt_lease_time
        self.jwt_audience = jwt_audience if len(jwt_audience) > 0 else None
        self.jwt_headers = {"kid": key_id} if key_id else None

        self.refresh_token_schema = RefreshTokenData.Schema()
        self.jwt_payload_data_schema = JWTPayloadData.Schema()

        self.async_session_factory = create_async_session_factory(ASYNC_DB_URI, Base)

    async def create_access_token(self, identifier: UUID) -> str:
        """
        Creates new JWT token.
        :param identifier: A UUID identifier of who the token was created to.
        :return: The new JWT token string
        """
        if self.private_key is None:
            raise RuntimeError("Private key required for JWT token creation")
        utc_now = datetime.utcnow()
        payload = JWTPayloadData(
            token_id=uuid4(),
            identifier=identifier,
            iat=int(utc_now.timestamp()),
            exp=int((utc_now + timedelta(seconds=self.jwt_lease_time)).timestamp()),
            aud=self.jwt_audience
        )
        encoded_payload = self.jwt_payload_data_schema.dump(payload)
        token_string = jwt.encode(encoded_payload, self.private_key, self.algorithm, headers=self.jwt_headers)
        return token_string


    async def load_access_token(self, token_string: str) -> JWTPayloadData:
        """
        Load + parse an existing JWT token.
        Raises exception if the token is incorrectly signed.
        """
        if not token_string:
            raise MissingJWTError
        token_dict: dict = jwt.decode(token_string, self.public_key, algorithms=[self.algorithm], options={"verify_exp": False})
        token_data = self.jwt_payload_data_schema.load(token_dict)
        return token_data


    async def create_refresh_token(self, key_id: int, expires_at: Optional[datetime] = None) -> str:
        """
        Create a new refresh token and insert to db.
        :param key_id: The id of the key the refresh token is created by.
        :return: The refresh token created.
        """
        if expires_at is None:
            expires_at = expires_at = datetime.utcnow() + timedelta(seconds=REFRESH_TOKEN_EXPIRY)
        if expires_at <= datetime.utcnow():
            raise RuntimeError("Refresh token can't be created in the past")

        refresh_token_str = sha512(uuid4().bytes).hexdigest()
        async with self.async_session_factory() as session:
            refresh_token_info_obj = RefreshTokenInfo(
                expires_at=expires_at,
                token=refresh_token_str,
                key_id=key_id
            )
            session.add(refresh_token_info_obj)
            await session.commit()
            await session.refresh(refresh_token_info_obj)
            await session.flush()
        return refresh_token_str


    async def check_refresh_token_exists(self, refresh_token: str) -> bool:
        """
        Check if a refresh token exists in DB.
        """
        async with self.async_session_factory() as session:
            query = select(func.count(RefreshTokenInfo.id)).where(RefreshTokenInfo.token == refresh_token)
            return (await session.scalar(query)) == 1


    async def load_refresh_token(self, refresh_token: str) -> RefreshTokenData:
        """
        Load a refresh token from DB.
        """
        if not await self.check_refresh_token_exists(refresh_token):
            raise InvalidTokenError("Invalid refresh token")
        async with self.async_session_factory() as session:
            query = select(RefreshTokenInfo).where(RefreshTokenInfo.token == refresh_token)
            refresh_token_info_obj = (await session.execute(query)).first()
            refresh_token_data_obj = self.refresh_token_schema.load(self.refresh_token_schema.dump(refresh_token_info_obj[0]))
            return refresh_token_data_obj



"""
A web server warpper for JWThenticatorAPI fucntions.
"""
from __future__ import absolute_import

from typing import Optional, Tuple
from http import HTTPStatus
from json.decoder import JSONDecodeError

from aiohttp import web
from marshmallow.exceptions import ValidationError
from jwt.exceptions import InvalidSignatureError, DecodeError

from jwthenticator import schemas, exceptions
from jwthenticator.api import JWThenticatorAPI
from jwthenticator.utils import get_rsa_key_pair
from jwthenticator.server_utils import extract_jwt
from jwthenticator.consts import PORT, URL_PREFIX, DISABLE_EXTERNAL_API, DISABLE_INTERNAL_API
from jwthenticator.loop_management import main_event_loop


class Server:

    # pylint: disable=too-many-arguments,too-many-instance-attributes
    def __init__(self, rsa_key_pair: Tuple[str, Optional[str]] = get_rsa_key_pair(), start_server: bool = True, port: int = PORT,
                 disable_external_api: bool = DISABLE_EXTERNAL_API, disable_internal_api: bool = DISABLE_INTERNAL_API):
        self.app = web.Application()
        self.api = JWThenticatorAPI(rsa_key_pair)

        self.jwt_validate_request_schema = schemas.JWTValidateRequest.Schema()
        self.bool_response_schema = schemas.BoolResponse.Schema()
        self.jkws_response_schema = schemas.JWKSResponse.Schema()
        self.auth_request_schema = schemas.AuthRequest.Schema()
        self.token_response_schema = schemas.TokenResponse.Schema()
        self.refresh_request_schema = schemas.RefreshRequest.Schema()
        self.key_request_schema = schemas.KeyRequest.Schema()
        self.register_key_request_schema = schemas.RegisterKeyRequest.Schema()

        # Disable certain exposed API functions. This is done to enable separating the service running
        #   between external and internal APIs (for example not to expose key registry externaly)
        if not disable_external_api:
            self.app.add_routes([
                web.post(f"{URL_PREFIX}/authenticate", self.authenticate),
                web.post(f"{URL_PREFIX}/refresh", self.refresh),
                web.post(f"{URL_PREFIX}/validate", self.validate),
                web.get(f"{URL_PREFIX}/validate_request", self.validate_request),
            ])

        if not disable_internal_api:
            self.app.add_routes([
                web.post(f"{URL_PREFIX}/register_key", self.register_key),
                web.post(f"{URL_PREFIX}/is_key_registered", self.is_key_registered),
            ])

        # General access routes
        self.app.add_routes([
            web.get(f"{URL_PREFIX}/", self.check_health),
            web.get(f"{URL_PREFIX}/health", self.check_health),
            web.get(f"{URL_PREFIX}/jwks", self.jwks),
        ])

        if start_server:
            web.run_app(self.app, port=port, loop=main_event_loop)


    async def authenticate(self, request: web.Request) -> web.Response:
        """
        :param request: AuthRequest
        :return: TokenResponse is successfull, BoleanResponse if not.
        """
        try:
            data = await request.json()
            authenticate_request = self.auth_request_schema.load(data)
            result = await self.api.authenticate(authenticate_request)
            return web.json_response(self.token_response_schema.dump(result), status=HTTPStatus.OK)
        except JSONDecodeError:
            return web.json_response({}, status=HTTPStatus.BAD_REQUEST)
        except ValidationError as err:
            return web.json_response(err.messages, status=HTTPStatus.BAD_REQUEST)
        except (exceptions.InvalidKeyError, exceptions.ExpiredError):
            return web.json_response({}, status=HTTPStatus.UNAUTHORIZED)


    async def refresh(self, request: web.Request) -> web.Response:
        """
        :param request: RefreshRequest
        :return: TokenResponse
        """
        try:
            data = await request.json()
            refresh_request = self.refresh_request_schema.load(data)
            result = await self.api.refresh(refresh_request)
            return web.json_response(self.token_response_schema.dump(result), status=HTTPStatus.OK)
        except JSONDecodeError:
            return web.json_response({}, status=HTTPStatus.BAD_REQUEST)
        except ValidationError as err:
            return web.json_response(err.messages, status=HTTPStatus.BAD_REQUEST)
        except (exceptions.InvalidTokenError, exceptions.ExpiredError):
            return web.json_response({}, status=HTTPStatus.UNAUTHORIZED)


    async def validate(self, request: web.Request) -> web.Response:
        """
        Validate a JWT (access) token.
        :param request: JWTValidateRequest
        :return: BoolResponse
        """
        try:
            data = await request.json()
            validate_request = self.jwt_validate_request_schema.load(data)
            result = await self.api.validate(validate_request)
            return web.json_response(self.bool_response_schema.dump(result), status=HTTPStatus.OK)
        except JSONDecodeError:
            return web.json_response({}, status=HTTPStatus.BAD_REQUEST)
        except ValidationError as err:
            return web.json_response(err.messages, status=HTTPStatus.BAD_REQUEST)
        except (InvalidSignatureError, DecodeError, exceptions.ExpiredError):
            return web.json_response({}, status=HTTPStatus.UNAUTHORIZED)


    async def validate_request(self, request: web.Request) -> web.Response:
        """
        Endpoint receives an empty request with authentication header,
            extracts JWT and validates it.
        This endpoint can be used with as an Nginx auth_request handler.
        :param request: Empty request (hopefully with authentication header)
        """
        try:
            jwt = await extract_jwt(request)
            validate_request = schemas.JWTValidateRequest(jwt)
            result = await self.api.validate(validate_request)
            if result.result:
                return web.json_response({}, status=HTTPStatus.OK)
            return web.json_response({}, status=HTTPStatus.UNAUTHORIZED)
        except (exceptions.MissingAuthorizationError, exceptions.BadAuthorizationError, exceptions.MissingJWTError):
            return web.json_response({}, status=HTTPStatus.FORBIDDEN)
        except exceptions.ExpiredError:
            return web.json_response({}, status=HTTPStatus.UNAUTHORIZED)


    async def register_key(self, request: web.Request) -> web.Response:
        """
        :param request: KeyRequest
        :return: BoolResponse
        """
        try:
            data = await request.json()
            key_request = self.register_key_request_schema.load(data)
            result = await self.api.register_key(key_request)
            return web.json_response(self.bool_response_schema.dump(result), status=HTTPStatus.CREATED)
        except JSONDecodeError:
            return web.json_response({}, status=HTTPStatus.BAD_REQUEST)
        except ValidationError as err:
            return web.json_response(err.messages, status=HTTPStatus.BAD_REQUEST)


    async def is_key_registered(self, request: web.Request) -> web.Response:
        """
        :param request: KeyRequest
        :return: BoolResponse
        """
        try:
            data = await request.json()
            key_request = self.key_request_schema.load(data)
            result = await self.api.is_key_registerd(key_request)
            return web.json_response(self.bool_response_schema.dump(result), status=HTTPStatus.OK)
        except JSONDecodeError:
            return web.json_response({}, status=HTTPStatus.BAD_REQUEST)
        except ValidationError as err:
            return web.json_response(err.messages, status=HTTPStatus.BAD_REQUEST)


    async def jwks(self, _request: web.Request) -> web.Response:
        """
        :return: JWKSResponse
        """
        result = await self.api.get_jwks()
        return web.json_response(self.jkws_response_schema.dump(result), status=HTTPStatus.OK)


    async def check_health(self, _request: web.Request) -> web.Response:
        """
        Health check method for the server, used to check if server is up and serving.
        :return: BoolResponse
        """
        result = schemas.BoolResponse(result=True)
        return web.json_response(self.bool_response_schema.dump(result), status=HTTPStatus.OK)


if __name__ == "__main__":
    Server()
